package hterhors.editor;

import hterhors.editor.format.DefaultFormattingStrategy;
import hterhors.editor.format.IgnoreFormattingStrategy;
import hterhors.editor.format.RuleFormattingStrategy;
import hterhors.editor.hoverinfos.GenerateHoverInfos;
import hterhors.editor.hoverinfos.IISRAnnontationHover;
import hterhors.editor.hyperlink.IsrHyperLinkDetector;
import hterhors.editor.scanners.ISRPartitionScanner;
import hterhors.editor.scanners.ISRSyntaxScanner;

import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ITextDoubleClickStrategy;
import org.eclipse.jface.text.ITextHover;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.text.TextAttribute;
import org.eclipse.jface.text.contentassist.ContentAssistEvent;
import org.eclipse.jface.text.contentassist.ContentAssistant;
import org.eclipse.jface.text.contentassist.ICompletionListener;
import org.eclipse.jface.text.contentassist.ICompletionProposal;
import org.eclipse.jface.text.contentassist.IContentAssistant;
import org.eclipse.jface.text.formatter.ContentFormatter;
import org.eclipse.jface.text.formatter.IContentFormatter;
import org.eclipse.jface.text.hyperlink.IHyperlinkDetector;
import org.eclipse.jface.text.presentation.IPresentationReconciler;
import org.eclipse.jface.text.presentation.PresentationReconciler;
import org.eclipse.jface.text.rules.DefaultDamagerRepairer;
import org.eclipse.jface.text.rules.Token;
import org.eclipse.jface.text.source.IAnnotationHover;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.SourceViewerConfiguration;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.texteditor.ITextEditor;

/**
 * @author  Hendrik
 */
public class ISRSourceViewerConfiguration extends SourceViewerConfiguration {

	/**
	 * @uml.property  name="doubleClickStrategy"
	 * @uml.associationEnd  
	 */
	private DoubleClickStrategy doubleClickStrategy;
	/**
	 * @uml.property  name="scanner"
	 * @uml.associationEnd  
	 */
	private ISRSyntaxScanner scanner;
	/**
	 * @uml.property  name="colorManager"
	 * @uml.associationEnd  
	 */
	private ColorManager colorManager;
	/**
	 * @uml.property  name="editor"
	 * @uml.associationEnd  
	 */
	private ISREditor editor;

	public ISRSourceViewerConfiguration(ColorManager colorManager,
			ISREditor editor) {
		this.colorManager = colorManager;
		this.editor = editor;
	}

	@Override
	public String[] getConfiguredContentTypes(ISourceViewer sourceViewer) {
		return new String[] { IDocument.DEFAULT_CONTENT_TYPE,
				ISRPartitionScanner.ISR_IGNORE, ISRPartitionScanner.ISR_RULE, };
	}

	@Override
	public ITextDoubleClickStrategy getDoubleClickStrategy(
			ISourceViewer sourceViewer, String contentType) {
		if (doubleClickStrategy == null)
			doubleClickStrategy = new DoubleClickStrategy();
		return doubleClickStrategy;
	}

	/**
	 * Returns the ISRSyntaxScanner. If scanner is null it will be created.
	 * 
	 * @return The current ISRSyntaxScanner
	 * 
	 */
	private ISRSyntaxScanner getISRSyntaxScanner() {
		if (scanner == null) {
			scanner = new ISRSyntaxScanner(colorManager);
			scanner.setDefaultReturnToken(new Token(new TextAttribute(
					colorManager.getColor(IISRColorConstants.DEFAULT))));
		}
		return scanner;
	}

	@Override
	public IPresentationReconciler getPresentationReconciler(
			ISourceViewer sourceViewer) {
		PresentationReconciler reconciler = new PresentationReconciler();
		DefaultDamagerRepairer dr = new DefaultDamagerRepairer(
				getISRSyntaxScanner());
		reconciler.setDamager(dr, IDocument.DEFAULT_CONTENT_TYPE);
		reconciler.setRepairer(dr, IDocument.DEFAULT_CONTENT_TYPE);

		dr = new DefaultDamagerRepairer(getISRSyntaxScanner());
		reconciler.setDamager(dr, ISRPartitionScanner.ISR_IGNORE);
		reconciler.setRepairer(dr, ISRPartitionScanner.ISR_IGNORE);
		dr = new DefaultDamagerRepairer(getISRSyntaxScanner());
		reconciler.setDamager(dr, ISRPartitionScanner.ISR_RULE);
		reconciler.setRepairer(dr, ISRPartitionScanner.ISR_RULE);
		dr = new DefaultDamagerRepairer(getISRSyntaxScanner());
		return reconciler;
	}

	@Override
	public IContentAssistant getContentAssistant(ISourceViewer sv) {
		ContentAssistant ca = new ContentAssistant();
		final ISRContentAssistantRuleProcessor carp = new ISRContentAssistantRuleProcessor();
		final ISRContentAssistantIgnoreRuleProcessor cairp = new ISRContentAssistantIgnoreRuleProcessor();
		final ISRContentAssistantDefaultProcessor cadp = new ISRContentAssistantDefaultProcessor();
		ca.setContentAssistProcessor(carp, ISRPartitionScanner.ISR_RULE);
		ca.setContentAssistProcessor(cairp, ISRPartitionScanner.ISR_IGNORE);
		ca.setContentAssistProcessor(cadp, IDocument.DEFAULT_CONTENT_TYPE);
		ca.setInformationControlCreator(getInformationControlCreator(sv));

		ca.addCompletionListener(new ICompletionListener() {

			@Override
			public void selectionChanged(ICompletionProposal arg0, boolean arg1) {
			}

			@Override
			public void assistSessionStarted(ContentAssistEvent arg0) {
				((ISRContentAssistantRuleProcessor) carp)
						.setOverride(getSelectedText(editor).length());
				((ISRContentAssistantIgnoreRuleProcessor) cairp)
						.setOverride(getSelectedText(editor).length());
				((ISRContentAssistantDefaultProcessor) cadp)
						.setOverride(getSelectedText(editor).length());

			}

			@Override
			public void assistSessionEnded(ContentAssistEvent arg0) {
			}
		});
		return ca;
	}

	/**
	 * Returns the current selected text from a specified editor.
	 * 
	 * @param editor
	 *            The Editor from which the selection should take.
	 * @return The current Selection
	 */
	private String getSelectedText(ITextEditor editor) {
		ISelection selection = editor.getSelectionProvider().getSelection();
		return ((ITextSelection) selection).getText();
	}

	@Override
	public IHyperlinkDetector[] getHyperlinkDetectors(ISourceViewer sourceViewer) {
		return new IHyperlinkDetector[] { new IsrHyperLinkDetector(editor) };
	}

	@Override
	public IContentFormatter getContentFormatter(ISourceViewer sourceViewer) {
		ContentFormatter formatter = new ContentFormatter();
		RuleFormattingStrategy ruleFormattingStrategy = new RuleFormattingStrategy();
		IgnoreFormattingStrategy ignoreFormattingStrategy = new IgnoreFormattingStrategy();
		DefaultFormattingStrategy defaultFormattingStrategy = new DefaultFormattingStrategy();
		formatter.setFormattingStrategy(defaultFormattingStrategy,
				IDocument.DEFAULT_CONTENT_TYPE);
		formatter.setFormattingStrategy(ignoreFormattingStrategy,
				ISRPartitionScanner.ISR_IGNORE);
		formatter.setFormattingStrategy(ruleFormattingStrategy,
				ISRPartitionScanner.ISR_RULE);
		return formatter;
	}

	@Override
	public ITextHover getTextHover(ISourceViewer sourceViewer,
			String contentType) {
		return new GenerateHoverInfos(sourceViewer, contentType);
	}

	@Override
	public IAnnotationHover getAnnotationHover(ISourceViewer sourceViewer) {
		super.getAnnotationHover(sourceViewer);
		return new IISRAnnontationHover();
	}

}